"""REST API and Server for the Experiment Manager."""

import datetime
from pathlib import Path
from typing import Callable, Optional

import uvicorn
from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.middleware.cors import CORSMiddleware
from madsci.client.event_client import EventClient, EventType
from madsci.common.ownership import ownership_context
from madsci.common.types.event_types import Event
from madsci.common.types.experiment_types import (
    Experiment,
    ExperimentManagerDefinition,
    ExperimentManagerSettings,
    ExperimentRegistration,
    ExperimentStatus,
)
from pymongo import MongoClient
from pymongo.database import Database


def create_experiment_server(  # noqa: C901, PLR0915
    experiment_manager_definition: Optional[ExperimentManagerDefinition] = None,
    experiment_manager_settings: Optional[ExperimentManagerSettings] = None,
    db_connection: Optional[Database] = None,
) -> FastAPI:
    """Creates an Experiment Manager's REST server."""
    logger = EventClient()
    experiment_manager_settings = (
        experiment_manager_settings or ExperimentManagerSettings()
    )
    logger.log_info(experiment_manager_settings)
    if not experiment_manager_definition:
        def_path = Path(
            experiment_manager_settings.experiment_manager_definition
        ).expanduser()
        if def_path.exists():
            experiment_manager_definition = ExperimentManagerDefinition.from_yaml(
                def_path,
            )
        else:
            experiment_manager_definition = ExperimentManagerDefinition()
        logger.log_info(f"Writing to experiment manager definition file: {def_path}")
        experiment_manager_definition.to_yaml(def_path)
    logger = EventClient(
        name=f"experiment_manager.{experiment_manager_definition.name}",
    )
    with ownership_context(
        manager_id=experiment_manager_definition.experiment_manager_id
    ):
        logger.log_info(experiment_manager_definition)

    # * DB Config
    if db_connection is None:
        db_client = MongoClient(experiment_manager_settings.db_url)
        db_connection = db_client["experiment_manager"]
    experiments = db_connection["experiments"]

    app = FastAPI()

    @app.middleware("http")
    async def ownership_middleware(request: Request, call_next: Callable) -> Response:
        """Middleware to set ownership context for each request."""
        with ownership_context(
            manager_id=experiment_manager_definition.experiment_manager_id
        ):
            return await call_next(request)

    @app.get("/")
    @app.get("/info")
    @app.get("/definition")
    async def definition() -> Optional[ExperimentManagerDefinition]:
        """Get the definition for the Experiment Manager."""
        return experiment_manager_definition

    @app.get("/experiment/{experiment_id}")
    async def get_experiment(experiment_id: str) -> Experiment:
        """Get an experiment by ID."""
        experiment = experiments.find_one({"_id": experiment_id})
        if not experiment:
            raise HTTPException(status_code=404, detail="Experiment not found")
        return Experiment.model_validate(experiment)

    @app.get("/experiments")
    async def get_experiments(number: int = 10) -> list[Experiment]:
        """Get the latest experiments."""
        experiments_list = (
            experiments.find().sort("started_at", -1).limit(number).to_list()
        )
        return [
            Experiment.model_validate(experiment) for experiment in experiments_list
        ]

    @app.post("/experiment")
    async def start_experiment(
        experiment_request: ExperimentRegistration,
    ) -> Experiment:
        """Start a new experiment."""
        experiment = Experiment.from_experiment_design(
            run_name=experiment_request.run_name,
            run_description=experiment_request.run_description,
            experiment_design=experiment_request.experiment_design,
        )
        experiment.started_at = datetime.datetime.now()
        experiments.insert_one(experiment.to_mongo())
        logger.log(
            event=Event(
                event_type=EventType.EXPERIMENT_START,
                event_data={"experiment": experiment},
            )
        )
        return experiment

    @app.post("/experiment/{experiment_id}/end")
    async def end_experiment(experiment_id: str) -> Experiment:
        """End an experiment by ID."""
        experiment = experiments.find_one({"_id": experiment_id})
        if not experiment:
            raise HTTPException(status_code=404, detail="Experiment not found")
        experiment = Experiment.model_validate(experiment)
        experiment.ended_at = datetime.datetime.now()
        experiment.status = ExperimentStatus.COMPLETED
        experiments.update_one(
            {"_id": experiment_id},
            {"$set": experiment.to_mongo()},
        )
        logger.log(
            event=Event(
                event_type=EventType.EXPERIMENT_COMPLETE,
                event_data={"experiment": experiment},
            )
        )
        return experiment

    @app.post("/experiment/{experiment_id}/continue")
    async def continue_experiment(experiment_id: str) -> Experiment:
        """Continue an experiment by ID."""
        experiment = experiments.find_one({"_id": experiment_id})
        if not experiment:
            raise HTTPException(status_code=404, detail="Experiment not found")
        experiment = Experiment.model_validate(experiment)
        experiment.status = ExperimentStatus.IN_PROGRESS
        experiments.update_one(
            {"_id": experiment_id},
            {"$set": experiment.to_mongo()},
        )
        logger.log(
            event=Event(
                event_type=EventType.EXPERIMENT_CONTINUED,
                event_data={"experiment": experiment},
            )
        )
        return experiment

    @app.post("/experiment/{experiment_id}/pause")
    async def pause_experiment(experiment_id: str) -> Experiment:
        """Pause an experiment by ID."""
        experiment = experiments.find_one({"_id": experiment_id})
        if not experiment:
            raise HTTPException(status_code=404, detail="Experiment not found")
        experiment = Experiment.model_validate(experiment)
        experiment.status = ExperimentStatus.PAUSED
        experiments.update_one(
            {"_id": experiment_id},
            {"$set": experiment.to_mongo()},
        )
        logger.log(
            event=Event(
                event_type=EventType.EXPERIMENT_PAUSE,
                event_data={"experiment": experiment},
            )
        )
        return experiment

    @app.post("/experiment/{experiment_id}/cancel")
    async def cancel_experiment(experiment_id: str) -> Experiment:
        """Cancel an experiment by ID."""
        experiment = experiments.find_one({"_id": experiment_id})
        if not experiment:
            raise HTTPException(status_code=404, detail="Experiment not found")
        experiment = Experiment.model_validate(experiment)
        experiment.status = ExperimentStatus.CANCELLED
        experiments.update_one(
            {"_id": experiment_id},
            {"$set": experiment.to_mongo()},
        )
        logger.log(
            event=Event(
                event_type=EventType.EXPERIMENT_CANCELLED,
                event_data={"experiment": experiment},
            )
        )
        return experiment

    @app.post("/experiment/{experiment_id}/fail")
    async def fail_experiment(experiment_id: str) -> Experiment:
        """Fail an experiment by ID."""
        experiment = experiments.find_one({"_id": experiment_id})
        if not experiment:
            raise HTTPException(status_code=404, detail="Experiment not found")
        experiment = Experiment.model_validate(experiment)
        experiment.status = ExperimentStatus.FAILED
        experiments.update_one(
            {"_id": experiment_id},
            {"$set": experiment.to_mongo()},
        )
        logger.log(
            event=Event(
                event_type=EventType.EXPERIMENT_FAILED,
                event_data={"experiment": experiment},
            )
        )
        return experiment

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    return app


if __name__ == "__main__":
    experiment_manager_settings = ExperimentManagerSettings()
    app = create_experiment_server(
        experiment_manager_settings=experiment_manager_settings,
    )
    uvicorn.run(
        app,
        host=experiment_manager_settings.experiment_server_url.host,
        port=experiment_manager_settings.experiment_server_url.port,
    )



"""
Test the Experiment Manager's REST server.

Uses pytest-mock-resources to create a MongoDB fixture. Note that this _requires_
a working docker installation.
"""

import pytest
from fastapi import status
from fastapi.testclient import TestClient
from madsci.common.types.experiment_types import (
    Experiment,
    ExperimentDesign,
    ExperimentManagerDefinition,
    ExperimentRegistration,
    ExperimentStatus,
)
from madsci.experiment_manager.experiment_server import create_experiment_server
from pymongo.database import Database
from pytest_mock_resources import MongoConfig, create_mongo_fixture

experiment_manager_def = ExperimentManagerDefinition(
    name="test_experiment_manager",
)


@pytest.fixture(scope="session")
def pmr_mongo_config() -> MongoConfig:
    """Congifure the MongoDB fixture."""
    return MongoConfig(image="mongo:8.0")


db_connection = create_mongo_fixture()


@pytest.fixture()
def test_client(db_connection: Database) -> TestClient:
    """Test client fixture for the Experiment Manager's server."""
    app = create_experiment_server(
        experiment_manager_definition=experiment_manager_def,
        db_connection=db_connection,
    )
    return TestClient(app)


def test_experiment_definition(test_client: TestClient) -> None:
    """
    Test the definition endpoint for the Experiment Manager's server.
    Should return an ExperimentManagerDefinition.
    """
    result = test_client.get("/definition").json()
    ExperimentManagerDefinition.model_validate(result)


def test_experiment_roundtrip(test_client: TestClient) -> None:
    """
    Test that we can send and then retrieve an experiment by ID.
    """
    test_experiment_design = ExperimentDesign(
        experiment_name="Test Experiment",
        experiment_description="This is a test experiment.",
    )
    payload = ExperimentRegistration(
        experiment_design=test_experiment_design,
        run_name="Test Run",
        run_description="This is a test run.",
    )
    response = test_client.post("/experiment", json=payload.model_dump(mode="json"))
    assert response.status_code == status.HTTP_200_OK
    assert response.json() is not None
    test_experiment = Experiment.model_validate(response.json())
    response = test_client.get(f"/experiment/{test_experiment.experiment_id}")
    assert response.status_code == status.HTTP_200_OK
    assert response.json() is not None
    assert Experiment.model_validate(response.json()) == test_experiment


def test_get_experiments(test_client: TestClient) -> None:
    """
    Test that we can retrieve all experiments and they are returned as a list in reverse-chronological order, with the correct number of experiments.
    """
    test_experiment_design = ExperimentDesign(
        experiment_name="Test Experiment",
        experiment_description="This is a test experiment.",
    )
    for i in range(10):
        payload = ExperimentRegistration(
            experiment_design=test_experiment_design,
            run_name=f"Test Experiment {i}",
            run_description=f"This is test experiment {i}.",
        )
        response = test_client.post(
            "/experiment",
            json=payload.model_dump(mode="json"),
        ).json()
        Experiment.model_validate(response)
    query_number = 5
    result = test_client.get("/experiments", params={"number": query_number}).json()
    # * Check that the number of experiments returned is correct
    assert len(result) == query_number
    previous_timestamp = float("inf")
    for experiment_data in result:
        experiment = Experiment.model_validate(experiment_data)
        # * Check that the experiments are in reverse-chronological order
        assert previous_timestamp >= experiment.started_at.timestamp()
        previous_timestamp = experiment.started_at.timestamp()


def test_end_experiment(test_client: TestClient) -> None:
    """
    Test that we can end an experiment by ID.
    """
    test_experiment_design = ExperimentDesign(
        experiment_name="Test Experiment",
        experiment_description="This is a test experiment.",
    )
    payload = ExperimentRegistration(
        experiment_design=test_experiment_design,
        run_name="Test Run",
        run_description="This is a test run.",
    )
    result = test_client.post(
        "/experiment", json=payload.model_dump(mode="json")
    ).json()
    test_experiment = Experiment.model_validate(result)
    result = test_client.post(f"/experiment/{test_experiment.experiment_id}/end").json()
    ended_experiment = Experiment.model_validate(result)
    assert ended_experiment.ended_at is not None
    assert ended_experiment.experiment_id == test_experiment.experiment_id


def test_cancel_experiment(test_client: TestClient) -> None:
    """
    Test that we can cancel an experiment by ID.
    """
    test_experiment_design = ExperimentDesign(
        experiment_name="Test Experiment",
        experiment_description="This is a test experiment.",
    )
    payload = ExperimentRegistration(
        experiment_design=test_experiment_design,
        run_name="Test Run",
        run_description="This is a test run.",
    )
    result = test_client.post(
        "/experiment", json=payload.model_dump(mode="json")
    ).json()
    test_experiment = Experiment.model_validate(result)
    result = test_client.post(
        f"/experiment/{test_experiment.experiment_id}/cancel"
    ).json()
    cancelled_experiment = Experiment.model_validate(result)
    assert cancelled_experiment.status == ExperimentStatus.CANCELLED
    assert cancelled_experiment.experiment_id == test_experiment.experiment_id


def test_pause_and_resume_experiment(test_client: TestClient) -> None:
    """
    Test that we can pause and resume an experiment by ID.
    """
    test_experiment_design = ExperimentDesign(
        experiment_name="Test Experiment",
        experiment_description="This is a test experiment.",
    )
    payload = ExperimentRegistration(
        experiment_design=test_experiment_design,
        run_name="Test Run",
        run_description="This is a test run.",
    )
    result = test_client.post(
        "/experiment", json=payload.model_dump(mode="json")
    ).json()
    test_experiment = Experiment.model_validate(result)
    result = test_client.post(
        f"/experiment/{test_experiment.experiment_id}/pause"
    ).json()
    paused_experiment = Experiment.model_validate(result)
    assert paused_experiment.status == ExperimentStatus.PAUSED
    assert paused_experiment.experiment_id == test_experiment.experiment_id
    result = test_client.post(
        f"/experiment/{test_experiment.experiment_id}/continue"
    ).json()
    resumed_experiment = Experiment.model_validate(result)
    assert resumed_experiment.status == ExperimentStatus.IN_PROGRESS
    assert resumed_experiment.experiment_id == test_experiment.experiment_id


def test_fail_experiment(test_client: TestClient) -> None:
    """
    Test that we can fail an experiment by ID.
    """
    test_experiment_design = ExperimentDesign(
        experiment_name="Test Experiment",
        experiment_description="This is a test experiment.",
    )
    payload = ExperimentRegistration(
        experiment_design=test_experiment_design,
        run_name="Test Run",
        run_description="This is a test run.",
    )
    result = test_client.post(
        "/experiment", json=payload.model_dump(mode="json")
    ).json()
    test_experiment = Experiment.model_validate(result)
    result = test_client.post(
        f"/experiment/{test_experiment.experiment_id}/fail"
    ).json()
    failed_experiment = Experiment.model_validate(result)
    assert failed_experiment.status == ExperimentStatus.FAILED
    assert failed_experiment.experiment_id == test_experiment.experiment_id
